import math
from collections import defaultdict
from collections import Counter
from typing import Dict, Generator, Iterable, List, Tuple, Union


from fuzzy_search.tokenization.string import score_levenshtein_distance
from fuzzy_search.tokenization.string import text2skipgrams
from fuzzy_search.tokenization.vocabulary import Vocabulary


def vector_length(skipgram_freq):
    return math.sqrt(sum([skipgram_freq[skip] ** 2 for skip in skipgram_freq]))


def get_skip_coocs(seq_ids: List[str], skip_size: int = 0) -> Generator[Tuple[int, int], None, None]:
    for ci, curr_id in enumerate(seq_ids):
        for next_id in seq_ids[ci + 1: ci + 2 + skip_size]:
            yield curr_id, next_id


def get_min_length(phrase1: str, phrase2: str, begin_length: int) -> int:
    return min([len(phrase1), len(phrase2), begin_length])


def get_begin_sim(phrase1: str, phrase2: str, begin_length: int) -> float:
    if min(len(phrase1), len(phrase2)) < begin_length:
        begin_length = min(len(phrase1), len(phrase2))
    begin1 = phrase1[:begin_length]
    begin2 = phrase2[:begin_length]
    return 1 - (score_levenshtein_distance(begin1, begin2) / begin_length)


def get_end_sim(phrase1: str, phrase2: str, end_length: int) -> float:
    if min(len(phrase1), len(phrase2)) < end_length:
        end_length = min(len(phrase1), len(phrase2))
    end1 = phrase1[-end_length:]
    end2 = phrase2[-end_length:]
    return 1 - (score_levenshtein_distance(end1, end2) / end_length)


class SkipCooccurrence:

    def __init__(self, vocabulary: Vocabulary, skip_size: int = 1, sentences: Iterable[List[str]] = None):
        """A class to count the co-occurrence frequency of word skipgrams."""
        self.cooc_freq = defaultdict(int)
        self.vocabulary = vocabulary
        self.skip_size: int = skip_size
        if sentences is not None:
            self.calculate_skip_cooccurrences(sentences)

    def calculate_skip_cooccurrences(self, sentences: Iterable[List[str]], skip_size: int = 0):
        """Count the frequency of term (skip) co-occurrences for a given list of sentences.

        :param sentences: a list of sentences, where each sentence is itself a list of term tokens
        :type sentences: Iterable[List[str]
        :param skip_size: the maximum number of skips to allow between co-occurring terms
        :type skip_size: int
        """
        for sent in sentences:
            seq_ids = [self.vocabulary.term2id(t) for t in sent]
            self.cooc_freq.update(get_skip_coocs(seq_ids, skip_size=skip_size))

    def _cooc_ids2terms(self, cooc_ids: Tuple[int, int]) -> Tuple[str, str]:
        id1, id2 = cooc_ids
        return self.vocabulary.id2term(id1), self.vocabulary.id2term(id2)

    def get_term_coocs(self, term: str) -> Union[None, Generator[Tuple[str, str], None, None]]:
        term_id = self.vocabulary.term2id(term)
        if term_id is None:
            return None
        for cooc_ids in self.cooc_freq:
            if term_id in cooc_ids:
                yield self._cooc_ids2terms(cooc_ids), self.cooc_freq[cooc_ids]


def is_close_distance_keyword_pair(keyword1: str, keyword2: str, max_distance_ratio: float,
                                   max_length_difference: int, max_distance: int) -> bool:
    if abs(len(keyword1) - len(keyword2)) > max_length_difference:
        return False
    distance = score_levenshtein_distance(keyword1, keyword2)
    if distance < max_distance and (
            distance / len(keyword1) < max_distance_ratio or distance / len(keyword2) < max_distance_ratio):
        return True
    return False


class KeywordList:

    def __init__(self, keywords: List[str], max_length_diff: int):
        self.len_keys = defaultdict(list)
        self.max_length_diff = max_length_diff
        for ki, keyword in enumerate(keywords):
            if isinstance(keyword, str) is False:
                raise ValueError(f"keyword '{keyword}' at index {ki} is not of "
                                 f"type str but type {type(keyword)}")
            len_key = len(keyword)
            self.len_keys[len_key].append(keyword)
        self.len_order = sorted(self.len_keys.keys())

    def iterate_candidate_pairs(self):
        for len_key1 in self.len_order:
            for ki, kw1 in enumerate(self.len_keys[len_key1]):
                for len_key2 in range(len_key1, len_key1 + self.max_length_diff + 1):
                    start = ki + 1 if len_key2 == len_key1 else 0
                    for kw2 in self.len_keys[len_key2][start:]:
                        yield kw1, kw2

    def find_close_distance_keywords(self, max_distance_ratio: float = 0.3,
                                     max_length_diff: int = 3, max_distance: int = 10,
                                     ignorecase: bool = False) -> Dict[str, List[str]]:
        """TODO: should we make the arguments into a config?"""
        if max_length_diff is None:
            max_length_diff = self.max_length_diff
        close_distance_keywords = defaultdict(list)
        for keyword1, keyword2 in self.iterate_candidate_pairs():
            string1 = keyword1.lower() if ignorecase else keyword1
            string2 = keyword2.lower() if ignorecase else keyword2
            if is_close_distance_keyword_pair(string1, string2, max_distance_ratio=max_distance_ratio,
                                              max_length_difference=max_length_diff,
                                              max_distance=max_distance):
                close_distance_keywords[keyword1].append(keyword2)
                close_distance_keywords[keyword2].append(keyword1)
        return close_distance_keywords

    def find_closer_terms(self, candidate: str, keyword: str, close_terms: List[str]):
        closer_terms = {}
        keyword_distance = score_levenshtein_distance(keyword, candidate)
        # print("candidate:", candidate, "\tkeyword:", keyword)
        # print("keyword_distance", keyword_distance)
        for close_term in close_terms:
            close_term_distance = score_levenshtein_distance(close_term, candidate)
            # print("close_term:", close_term, "\tdistance:", close_term_distance)
            if close_term_distance < keyword_distance:
                closer_terms[close_term] = close_term_distance


class SkipgramSimilarity:

    def __init__(self, ngram_length: int = 3, skip_length: int = 0, terms: List[str] = None,
                 max_length_diff: int = 2):
        """A class to index terms by their character skipgrams and to find similar terms for a given
        input term based on the cosine similarity of their skipgram overlap.

        :param ngram_length: the number of characters per ngram
        :type ngram_length: int
        :param skip_length: the maximum number of characters to skip
        :type skip_length: int
        :param terms: a list of terms
        :type terms: List[str]
        :param max_length_diff: the maximum difference in length between a search term and a term in the
        index to be considered a match. This is an efficiency parameter to reduce the number of candidate
        similar terms to ones that are roughly similar in length to the search term.
        :type max_length_diff: int
        :
        """
        self.ngram_length = ngram_length
        self.skip_length = skip_length
        self.vocabulary = Vocabulary()
        self.vector_length = {}
        self.max_length_diff = max_length_diff
        self.skipgram_index = defaultdict(lambda: defaultdict(Counter))
        if terms is not None:
            self.index_terms(terms)

    def _reset_index(self):
        self.vocabulary.reset_index()
        self.vector_length = {}
        self.skipgram_index = defaultdict(lambda: defaultdict(Counter))

    def index_terms(self, terms: List[str], reset_index: bool = True):
        """Make a frequency index of the skip grams for a given list of terms.
        By default, indexing is cumulative, that is, everytime you call index_terms
        with a list of terms, they are added to the index. Use 'reset_index=True' to
        reset the index before indexing the given terms.

        :param terms: a list of term to index
        :type terms: List[str]
        :param reset_index: whether to reset the index before indexing or to keep the existing index
        :type reset_index: bool
        """
        if reset_index is True:
            self._reset_index()
        self.vocabulary.add_terms(terms)
        for term in terms:
            self._index_term_skips(term)

    def _term_to_skip(self, term):
        skip_gen = text2skipgrams(term, ngram_size=self.ngram_length, skip_size=self.skip_length)
        return Counter([skip.string for skip in skip_gen])

    def _index_term_skips(self, term: str):
        term_id = self.vocabulary.term_id[term]
        skipgram_freq = self._term_to_skip(term)
        self.vector_length[term_id] = vector_length(skipgram_freq)
        for skipgram in skipgram_freq:
            # print(skip.string)
            self.skipgram_index[skipgram][len(term)][term_id] = skipgram_freq[skipgram]

    def _get_term_vector_length(self, term, skipgram_freq):
        if term not in self.vocabulary.term_id:
            return vector_length(skipgram_freq)
        else:
            term_id = self.vocabulary.term_id[term]
            return self.vector_length[term_id]

    def _compute_dot_product(self, term):
        skipgram_freq = self._term_to_skip(term)
        term_vl = self._get_term_vector_length(term, skipgram_freq)
        # print(term, 'vl:', term_vl)
        dot_product = defaultdict(int)
        for skipgram in skipgram_freq:
            for term_length in range(len(term) - self.max_length_diff, len(term) + self.max_length_diff + 1):
                for term_id in self.skipgram_index[skipgram][term_length]:
                    dot_product[term_id] += skipgram_freq[skipgram] * self.skipgram_index[skipgram][term_length][
                        term_id]
                    # print(term_id, self.vocab_map[term_id], dot_product[term_id])
        for term_id in dot_product:
            dot_product[term_id] = dot_product[term_id] / (term_vl * self.vector_length[term_id])
        return dot_product

    def rank_similar(self, term: str, top_n: int = 10, score_cutoff: float = 0.5):
        """Return a ranked list of similar terms from the index for a given input term,
        based on their character skipgram cosine similarity.

        :param term: a term (any string) to match against the indexed terms
        :type term: str
        :param top_n: the number of highest ranked terms to return
        :type top_n: int (default 10)
        :param score_cutoff: the minimum similarity score after which to cutoff the ranking
        :type score_cutoff: float
        :return: a ranked list of terms and their similarity scores
        :rtype: List[Tuple[str, float]]
        """
        dot_product = self._compute_dot_product(term)
        top_terms = []
        for term_id in sorted(dot_product, key=lambda t: dot_product[t], reverse=True):
            if dot_product[term_id] < score_cutoff:
                break
            term = self.vocabulary.id_term[term_id]
            top_terms.append((term, dot_product[term_id]))
            if len(top_terms) == top_n:
                break
        return top_terms



from __future__ import annotations

import copy
import time
from collections import defaultdict
from typing import Dict, List, Union

from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.match.phrase_match import PhraseMatch
from fuzzy_search.match.phrase_match import MatchType
from fuzzy_search.match.phrase_match import TokenMatch
from fuzzy_search.match.phrase_match import PartialPhraseMatch
from fuzzy_search.match.phrase_match import copy_partial_match
from fuzzy_search.match.skip_match import SkipMatches
from fuzzy_search.phrase.phrase_model import PhraseModel
from fuzzy_search.search.searcher import FuzzySearcher
from fuzzy_search.tokenization.string import SkipGram
from fuzzy_search.tokenization.string import score_levenshtein_similarity_ratio
from fuzzy_search.tokenization.string import text2skipgrams
from fuzzy_search.tokenization.token import Doc
from fuzzy_search.tokenization.token import Token
from fuzzy_search.tokenization.token import Tokenizer
from fuzzy_search.tokenization.vocabulary import Vocabulary


def map_text_tokens_to_phrase_tokens(partial_match: PartialPhraseMatch) -> Union[Dict[str, List[str]], None]:
    # check that each text token appears only once
    # if a text token appears multiple times, pick its best representation
    # compute the per token levenshtein similarity
    # compute the all tokens levenshtein similarity
    # compute the whole string levenshtein similarity
    text_phrase_map = defaultdict(list)
    phrase_token_set = set()
    phrase_tokens = [token.n for token in partial_match.phrase.tokens]
    for token_match in partial_match.token_matches:
        for phrase_token in token_match.phrase_tokens:
            phrase_token_set.add(phrase_token)
            for text_token in token_match.text_tokens:
                text_phrase_map[text_token].append(phrase_token)
                if phrase_token in phrase_tokens:
                    phrase_tokens.remove(phrase_token)
    if len(phrase_tokens) > 0:
        return None
    return text_phrase_map


def get_partial_phrases(token_matches: List[TokenMatch], token_searcher: FuzzyTokenSearcher,
                        max_char_gap: int = 20, debug: int = 0):
    partial_phrase: Dict[Phrase, PartialPhraseMatch] = {}
    candidate_phrase: Dict[Phrase, List[PartialPhraseMatch]] = defaultdict(list)
    prev_token_match = None
    if debug > 1:
        print(f"token_searcher.get_partial_phrases")
    for token_match in token_matches:
        if prev_token_match and token_match.text_start - prev_token_match.text_end > max_char_gap:
            if debug > 1:
                gap = token_match.text_start - prev_token_match.text_end
                print(f"    gap between prev_token_match and token_match ({gap}) "
                      f"bigger than max_char_gap {max_char_gap}")
                print(f"      emptying partial_phrase {partial_phrase}")
            partial_phrase: Dict[Phrase, PartialPhraseMatch] = {}
        if debug > 1:
            print(
                f"text start: {token_match.text_start: >4}\tend{token_match.text_end: >4}\t\t"
                f"text tokens: {token_match.text_tokens}\t\tphrase tokens{token_match.phrase_tokens}")
        for phrase_token_string in token_match.phrase_tokens:
            if debug > 1:
                print('\tphrase_token_string:', phrase_token_string)
            for phrase_string in token_searcher.phrase_model.token_in_phrase[phrase_token_string]:
                # phrase = token_searcher.phrase_model.phrase_index[phrase_string]
                phrase = token_searcher.phrase_model.get_phrase(phrase_string)
                if phrase is None:
                    continue
                # print('phrase:', phrase)
                # print('phrase.tokens:', phrase.tokens)
                if debug > 1 and phrase is not None:
                    print('\t\tphrase:', phrase)
                if phrase not in candidate_phrase:
                    if debug > 1:
                        print('\t\t\tADDING PHRASE AND APPENDING PARTIAL')
                    partial = PartialPhraseMatch(phrase, [token_match])
                    # partial_phrase[phrase] = partial
                    candidate_phrase[phrase].append(partial)
                else:
                    added = False
                    for partial in candidate_phrase[phrase]:
                        # check if token extends existing partial match
                        if token_match.text_start - partial.text_end > max_char_gap:
                            # print('skipping partial match:', token_match.text_start, partial.text_end)
                            continue
                        if debug > 1:
                            print(f"TEST partial.text_end: {partial.text_end} "
                                  f"token_match.text_start: {token_match.text_start}")
                            print(f"    partial.missing_tokens: {partial.missing_tokens}")
                            print(f"    token_match.phrase_tokens: {token_match.phrase_tokens}")
                        if partial.text_end < token_match.text_start and \
                                any([pt for pt in token_match.phrase_tokens if pt in partial.missing_tokens]):
                            partial_copy = copy_partial_match(partial)
                            partial_copy.add_tokens([token_match])
                            if partial_copy.text_length - len(phrase) <= token_searcher.config['max_length_variance']:
                                candidate_phrase[phrase].append(partial_copy)
                                if debug > 1:
                                    print('\t\t\tADDING TOKEN TO COPY OF PARTIAL')
                                    print('\t\t', partial_copy.text_tokens)
                                    print('\t\t', partial_copy.text_start, partial_copy.text_end)
                                    print('\t\t', partial_copy.text_length, len(phrase))
                                added = True
                    if not added:
                        partial = PartialPhraseMatch(phrase, [token_match])
                        if debug > 1:
                            print('\t\t\tADDING AS NEW PARTIAL TO PHRASE')
                            print('\t\t', partial.text_tokens)
                            print('\t\t', partial.text_start, partial.text_end)
                            print('\t\t', partial.text_length, len(phrase))
                        candidate_phrase[phrase].append(partial)
            if debug > 1:
                print(f"\n--------------\nCANDIDATE PHRASES:")
                for phrase in candidate_phrase:
                    print(f"    PHRASE: {phrase}")
                    for partial in candidate_phrase[phrase]:
                        print(f"\tPARTIAL: {partial.text_tokens}, {partial.phrase_tokens}\t "
                              f"RANGE: {partial.text_start}-{partial.text_end}\t"
                              f"MISSING: {partial.missing_tokens}")
                print('---------------')
        prev_token_match = token_match
    for phrase in candidate_phrase:
        remove_partials = []
        remove_incomplete = False
        if any([len(partial.missing_tokens) == 0 for partial in candidate_phrase[phrase]]):
            remove_incomplete = True
        for partial in candidate_phrase[phrase]:
            if debug > 1:
                print('\tCHECKING PARTIAL CANDIDATE')
                print('\t\t', partial.text_tokens)
                print('\t\t', partial.text_start, partial.text_end)
                print('\t\t', partial.text_length, len(phrase))
            if remove_incomplete is True and len(partial.missing_tokens) > 0:
                if debug > 1:
                    print("\t\tREMOVE CANDIDATE BECAUSE IT'S INCOMPLETE!")
                remove_partials.append(partial)
                # candidate_phrase[phrase].remove(partial)
            elif abs(partial.text_length - len(phrase)) > token_searcher.config['max_length_variance']:
                if debug > 0:
                    print('\t\tREMOVE CANDIDATE!')
                remove_partials.append(partial)
                # candidate_phrase[phrase].remove(partial)
                # candidate_phrase.append(partial)
                # del partial
            # elif partial.text_length > len(phrase) + token_searcher.config['max_length_variance']:
            else:
                if debug > 0:
                    print('\t\tCANDIDATE FOUND!')
                # del partial[phrase]
        for partial in remove_partials:
            candidate_phrase[phrase].remove(partial)
    return candidate_phrase


def get_tokenized_doc(text: Union[str, Dict[str, any], Doc], tokenizer: Tokenizer) -> Doc:
    if isinstance(text, Doc):
        return text
    elif isinstance(text, dict):
        return tokenizer.tokenize_doc(text['text'], doc_id=text['id'])
    elif isinstance(text, str):
        return tokenizer.tokenize_doc(text)
    else:
        raise TypeError(f"text must be str, dict (with 'text' and 'id' properties) or Doc, not {type(text)}")


def get_text_tokens(text: Union[str, Dict[str, any], Doc], tokenizer: Tokenizer = None):
    if isinstance(text, Doc):
        return text.tokens
    elif isinstance(text, list) and all(isinstance(ele, Token) for ele in text):
        return text
    elif isinstance(text, str):
        return tokenizer.tokenize(doc_text=text)
    elif isinstance(text, dict):
        return tokenizer.tokenize(doc_text=text['text'], doc_id=text['id'])
    else:
        raise TypeError(
            f'invalid text type {type(text)}, must be string, Doc or a dictionary with "text" and "id" properties')


def get_text_string(text: Union[str, Dict[str, any], Doc]) -> str:
    if isinstance(text, Doc):
        return text.text
    elif isinstance(text, list) and all(isinstance(ele, Token) for ele in text):
        return ' '.join([token.normalised_string for token in text])
    elif isinstance(text, str):
        return text
    elif isinstance(text, dict):
        return text['text']
    else:
        raise TypeError(
            f'invalid text type {type(text)}, must be string, Doc or a dictionary with "text" and "id" properties')


def get_token_skipgram_matches(text_token: Token, text_token_skips: List[SkipGram],
                               token_searcher: FuzzyTokenSearcher,
                               debug: int = 0):
    token_skip_matches = SkipMatches(token_searcher.ngram_size, token_searcher.skip_size)
    for skipgram in text_token_skips:
        for phrase_token in token_searcher.token_skipgram_index[skipgram.string]:
            if phrase_token in token_searcher.phrase_model.phrase_token_max_start_offset:
                max_token_offset = token_searcher.phrase_model.phrase_token_max_start_offset[phrase_token]
                if text_token.char_index > max_token_offset:
                    continue
            token_skip_matches.add_skip_match(skipgram, phrase_token)
    return token_skip_matches


def has_max_start_offset(phrase: Phrase):
    return phrase.max_start_offset is not None and phrase.max_start_offset != -1


def has_max_end_offset(phrase: Phrase):
    return phrase.max_end_offset is not None and phrase.max_end_offset != -1


class FuzzyTokenSearcher(FuzzySearcher):

    def __init__(self, phrase_list: List[any] = None,
                 phrase_model: Union[Dict[str, any], List[Dict[str, any]], PhraseModel] = None,
                 config: Union[None, Dict[str, Union[str, int, float]]] = None, tokenizer: Tokenizer = None,
                 vocabulary: Vocabulary = None, max_char_gap: int = 20, max_token_gap: int = 1):
        """This class represents the basic fuzzy searcher. You can pass a list of phrases or a phrase model and
        configuration dictionary that overrides the default configuration values. The default config dictionary
        is available via `fuzzy_search.default_config`.

        To set e.g. the character ngram_size to 3 and the skip_size to 1 use the following dictionary:

        config = {
            'ngram_size': 3,
            'skip_size': 1
        }

        :param phrase_list: a list of phrases (a list of strings or more complex dictionaries with phrases and variants)
        :type phrase_list: list
        :param phrase_model: a phrase model
        :type phrase_model: PhraseModel
        :param config: a configuration dictionary to override default configuration properties.
        Only the properties in the config dictionaries of updated.
        :type config: dict
        :param tokenizer: a tokenizer instance
        :type tokenizer: Tokenizer
        """
        super().__init__(phrase_list=phrase_list, phrase_model=phrase_model,
                         config=config, tokenizer=tokenizer)
        debug = self.config['debug']
        self.token_skipgram_index = defaultdict(set)
        self.token_num_skips = {}
        self.max_token_gap = max_token_gap
        self.max_char_gap = max_char_gap
        if 'max_token_length_variance' not in self.config:
            self.config['max_token_length_variance'] = self.config['max_length_variance']
        self.vocabulary = vocabulary
        if debug > 3:
            print(f'{self.__class__.__name__}.index_phrase_model - calling index_token_skipgrams()')
        if debug > 3:
            print(f'{self.__class__.__name__}.index_phrase_model - calling index_token_skipgrams()')
        if self.phrase_model is not None:
            if debug > 3:
                print(f'{self.__class__.__name__}.index_phrase_model - calling index_token_skipgrams()')
            self.index_token_skipgrams()

    def configure(self, config: Dict[str, any]):
        for prop in config:
            if hasattr(self, prop):
                self.__setattr__(prop, config[prop])

    def index_token_skipgrams(self, debug: int = 0):
        debug = self._get_debug_level(debug)
        for token_string in self.phrase_model.token_in_phrase:
            if debug > 2:
                print(f'\tindex_token_skipgrams - token_string: {token_string}')
            skips = [skip for skip in text2skipgrams(token_string, self.ngram_size, skip_size=self.skip_size)]
            self.token_num_skips[token_string] = len(skips)
            for skipgram in skips:
                self.token_skipgram_index[skipgram.string].add(token_string)

    def find_skipgram_token_matches_in_text(self, text: Union[Doc, str, Dict[str, any]],
                                            debug: int = 0) -> List[TokenMatch]:
        """Find all token matches between text tokens and phrase tokens using skipgrams.

        :param text: the text object to match with tokens
        :type text: Dict[str, Union[str, int, float, list]]
        :param debug: level to show debug information
        :type debug: int
        :return: a list of matches between text tokens and phrase tokens
        :rtype: List[TokenMatch]
        """
        text_tokens = get_text_tokens(text, self.tokenizer)
        token_matches = []
        partial_matches = defaultdict(list)
        for text_token in text_tokens:
            if self.vocabulary and self.vocabulary.has_term(text_token):
                self.find_vocabulary_token_matches_for_token(text_token, token_matches, debug=debug)
                continue
            if debug > 0:
                print('\n  find_skipgram_token_matches_in_text - text_token:', text_token)
            self.find_skipgram_token_matches_for_token(text_token, partial_matches, token_matches, debug=debug)
        return token_matches

    def find_vocabulary_token_matches_for_token(self, text_token: Token, token_matches: List[TokenMatch],
                                                debug: int = 0):
        if text_token.n not in self.phrase_model.token_in_phrase:
            return None
        for phrase_string in self.phrase_model.token_in_phrase[text_token.n]:
            phrase = self.phrase_model.get_phrase(phrase_string)
            if isinstance(phrase, Phrase):
                if token_is_out_of_phrase_range(text_token, phrase, self):
                    continue
                phrase_tokens = [token for token in phrase.tokens if token.n == text_token.n]
                token_match = TokenMatch(text_tokens=text_token, phrase_tokens=phrase_tokens,
                                         match_type=MatchType.FULL)
                token_matches.append(token_match)

    def find_skipgram_token_matches_for_token(self, text_token: Token, partial_matches: Dict[str, List[Token]],
                                              token_matches: List[TokenMatch], debug: int = 0):
        """Find all token matches between text tokens and phrase tokens using skipgrams.

        :param text_token: a single text token to match with phrase tokens
        :type text_token: Token
        :param partial_matches: a dictionary of phrase token strings an their partial text token matches
        :type partial_matches: Dict[str, List[Token]]
        :param token_matches: a list of matches between text tokens and phrase tokens
        :type token_matches: List[TokenMatch]
        :param debug: level to show debug information
        :type debug: int
        """
        if debug > 0:
            print(f'\n    find_skipgram_token_matches_for_token - text_token: {text_token}')
        text_token_skips = [skipgram for skipgram in text2skipgrams(text_token.n, self.ngram_size,
                                                                    self.skip_size)]
        if debug > 0:
            print(f'\n    find_skipgram_token_matches_for_token - text_token skips: {text_token_skips}')
        token_skip_matches = get_token_skipgram_matches(text_token, text_token_skips, self)
        if debug > 0:
            print(f'    find_skipgram_token_matches_for_token - number of phrase matches: '
                  f'{len(token_skip_matches.match_start_offsets)}')
            print()
        text_token_num_skips = len(text_token_skips)
        token_match_types = set()
        for pmi, phrase_token_match in enumerate(token_skip_matches.match_start_offsets):
            if debug > 1:
                print(f'\n    find_skipgram_token_matches_for_token - phrase_token_match {pmi+1}: {phrase_token_match}')
                print(f'    find_skipgram_token_matches_for_token - number of skipgram matches: '
                      f'{len(token_skip_matches.match_start_offsets[phrase_token_match])}')
                print(f'    find_skipgram_token_matches_for_token - skipgram matches: ',
                      token_skip_matches.match_skipgrams[phrase_token_match][0].start_offset,
                      token_skip_matches.match_skipgrams[phrase_token_match][-1].end_offset,
                      )
            match_type = get_token_skip_match_type(text_token.normalised_string,
                                                   text_token_num_skips,
                                                   token_skip_matches,
                                                   phrase_token_match, self, debug=debug)
            if debug > 2:
                print(f'    find_skipgram_token_matches - match_type: {match_type}')
            token_match_types.add(match_type)
            if match_type == MatchType.NONE:
                continue
            elif match_type == MatchType.FULL:
                token_match = TokenMatch(text_token, phrase_token_match, match_type)
                if debug > 2:
                    print(f'\n    find_skipgram_token_matches - adding full match: {token_match}\n')
                token_matches.append(token_match)
                if debug > 0:
                    print('NUMBER OF TOKEN MATCHES:', len(token_matches))
            elif match_type == MatchType.PARTIAL_OF_PHRASE_TOKEN:
                if debug > 2:
                    print(f'\tfind_skipgram_token_matches - partial_matches: {partial_matches}')
                if phrase_token_match in partial_matches:
                    if debug > 2:
                        print(phrase_token_match)
                        print(partial_matches)
                        print(partial_matches[phrase_token_match])
                    last_partial = partial_matches[phrase_token_match][-1]
                    if debug > 2:
                        print(f'\tfind_skipgram_token_matches - last_partial: {last_partial}')
                    if text_token.char_index - (last_partial.char_index + len(last_partial)) > 4:
                        if debug > 2:
                            print(f'\t\tfind_skipgram_token_matches - text_token.char_index: {text_token.char_index}')
                            print(f'\t\tfind_skipgram_token_matches - end of last_partial: '
                                  f'{last_partial.char_index + len(last_partial)}')
                            print(f'\t\tfind_skipgram_token_matches - removing partial_match for phrase token: '
                                  f'{phrase_token_match}')
                        del partial_matches[phrase_token_match]
                if debug > 2:
                    print(f'\tfind_skipgram_token_matches - adding partial "{text_token}" '
                          f'of phrase token "{phrase_token_match}"')
                partial_matches[phrase_token_match].append(text_token)
                if len(partial_matches[phrase_token_match]) > 1:
                    if debug > 2:
                        print(f'\t\tfind_skipgram_token_matches - checking if partial_matches align with '
                              f'phrase token: {partial_matches[phrase_token_match]}')
                    first_partial = partial_matches[phrase_token_match][0]
                    last_partial = partial_matches[phrase_token_match][-1]
                    partial_length = last_partial.char_index + len(last_partial) - first_partial.char_index
                    length_diff = partial_length - len(phrase_token_match)
                    if debug > 2:
                        print(f'\t\tfind_skipgram_token_matches - '
                              f'partial_matches: {partial_matches[phrase_token_match]}')
                        print(f'\t\tfind_skipgram_token_matches - partial_length: {partial_length}')
                        print(f'\t\tfind_skipgram_token_matches - phrase_token length: {phrase_token_match}')
                        print(f'\t\tfind_skipgram_token_matches - lenght_diff: {length_diff}')

                    if length_diff > 0 or abs(length_diff) <= self.config['max_token_length_variance']:
                        token_match = TokenMatch(partial_matches[phrase_token_match],
                                                 phrase_token_match, match_type)
                        token_matches.append(token_match)
                        if debug > 0:
                            print('NUMBER OF TOKEN MATCHES:', len(token_matches))
                        if debug > 2:
                            print(f'\n    find_skipgram_token_matches - adding full match: {token_match}')
                        partial_matches[phrase_token_match].pop(0)
            elif match_type == MatchType.PARTIAL_OF_TEXT_TOKEN:
                token_match = TokenMatch(text_token, phrase_token_match, match_type)
                token_matches.append(token_match)
        if MatchType.PARTIAL_OF_PHRASE_TOKEN not in token_match_types:
            if debug > 2:
                print(f'    find_skipgram_token_matches - emptying partial_matches')
            partial_matches = defaultdict(list)
        if debug > 1:
            print(partial_matches, '\n')

    def _pick_best_candidates(self, doc: Doc, candidate_phrases: Dict[Phrase, List[PartialPhraseMatch]],
                              debug: int = 0):
        if debug > 0:
            print('\n  _pick_best_candidates - listing partial matches:')
            for phrase in candidate_phrases:
                print('    _pick_best_candidates - phrase:', phrase)
                for partial in candidate_phrases[phrase]:
                    print('\t_pick_best_candidates - partial:', partial)
        phrase_at_offset: Dict[int, PartialPhraseMatch] = {}
        for phrase in candidate_phrases:
            for pp in candidate_phrases[phrase]:
                full_match_string = doc.text[pp.text_start:pp.text_end]
                full_text_length = pp.text_end - pp.text_start
                # the match_string is only the text tokens, not the text between
                # start and end offsets: if there are non-matching tokens in between,
                # the Levenshtein score should be computed only on the matching text
                # tokens.
                pp.match_string = ' '.join([token.n for token in pp.text_tokens])
                if debug > 0:
                    print('\n    _pick_best_candidates - pp:', pp.phrase.phrase_string)
                    print('\t_pick_best_candidates - pp.match_string:', pp.match_string)
                if abs(len(pp.match_string) - len(pp.phrase.phrase_string)) > self.max_length_variance:
                    if abs(full_text_length - len(pp.phrase.phrase_string)) > self.max_length_variance:
                        if debug > 0:
                            print(f"      length difference above max length variance")
                        continue
                    else:
                        if debug > 0:
                            print(f"      length difference above max length variance, but full text length below")
                pp.levenshtein_score = score_levenshtein_similarity_ratio(pp.phrase.phrase_string, pp.match_string)
                if debug > 1:
                    print('\t_pick_best_candidates - pp.levenshtein_score:', pp.levenshtein_score)
                if pp.levenshtein_score < self.config['levenshtein_threshold']:
                    if debug > 0:
                        print(f'\t\t_pick_best_candidates - phrase below '
                              f'score threshold {pp.text_start}:', pp.phrase.phrase_string)
                    continue
                if pp.text_start in phrase_at_offset:
                    if debug > 0:
                        print(f'\t\t_pick_best_candidates - next phrase at '
                              f'offset {pp.text_start}:', pp.phrase.phrase_string)
                    if pp.levenshtein_score > phrase_at_offset[pp.text_start].levenshtein_score:
                        if debug > 0:
                            print(f'\t\t_pick_best_candidates - replacing '
                                  f'phrase at offset {pp.text_start}:', pp.phrase.phrase_string)
                        phrase_at_offset[pp.text_start] = pp
                else:
                    if debug > 0:
                        print(f'\t\t_pick_best_candidates - first phrase at '
                              f'offset {pp.text_start}:', pp.phrase.phrase_string)
                    phrase_at_offset[pp.text_start] = pp
        phrases = []
        for partial_phrase in sorted(phrase_at_offset.values(), key=lambda x: x.text_start):
            if partial_phrase.phrase.phrase_string in self.phrase_model.phrase_index:
                phrase = partial_phrase.phrase
                variant = phrase
            elif partial_phrase.phrase.phrase_string in self.phrase_model.variant_index:
                phrase = self.phrase_model.variant_of(partial_phrase.phrase)
                variant = partial_phrase.phrase
            elif partial_phrase.phrase.phrase_string in self.phrase_model.distractor_index:
                # phrase is a distractor
                continue
            else:
                raise KeyError(f"partial phrase {partial_phrase.phrase} not registered in phrase_model.")
            phrase = PhraseMatch(phrase, variant, partial_phrase.match_string,
                                 partial_phrase.text_start, text_id=doc.id,
                                 levenshtein_similarity=partial_phrase.levenshtein_score,
                                 match_label=list(partial_phrase.phrase.label_set))
            phrases.append(phrase)
        return phrases

    def find_matches(self, text: Union[Doc, str, Dict[str, any]],
                     debug: int = 0) -> List[PhraseMatch]:
        """Find all fuzzy matching phrases for a given text using token-based searching.
        The `FuzzyTokenSearcher` turns the phrases and the target text into lists of word
        tokens (the tokenizer is configurable) and uses character skip grams to identify
        candidate phrase tokens matching tokens in the text. It then uses token sequences
        to identify fuzzy matches.

        This speeds up the search (especially for the default settings `ngram_size=2` and
        `skip_size=2`) at the cost of slightly less exhaustive search.

        :param text: A tokenized text.
        :type text: Union[str, Dict[str, str]]
        :param debug: level to show debug information
        :type debug: int
        :return: a list of phrase matches
        :rtype: List[PartialPhraseMatch]
        """
        start = time.time()
        doc = get_tokenized_doc(text, self.tokenizer)
        if debug > 0:
            print('find_matches - checking token matches in text')
        token_matches = self.find_skipgram_token_matches_in_text(doc, debug=debug)
        if debug > 0:
            print('find_matches - number of token_matches:', len(token_matches))
        if debug > 1:
            step1 = time.time()
            print(f'step 1 took: {step1 - start: >.2f} seconds')
        if debug > 1:
            for tm in token_matches:
                print('find_matches - token_match:', tm)
        candidate_phrases = get_partial_phrases(token_matches, self, debug=debug)
        if debug > 1:
            print('find_matches - number of candidate phrases:', len(candidate_phrases))
            print('find_matches - number of partial phrases:',
                  sum([len(candidate_phrases[phrase]) for phrase in candidate_phrases]))
            step2 = time.time()
            print(f'step 2 took: {step2 - step1: >.2f} seconds')
        matches = self._pick_best_candidates(doc, candidate_phrases, debug=debug)
        filtered_matches = self.filter_matches_by_offset_threshold(matches)
        if debug > 1:
            step3 = time.time()
            print(f'step 3 took: {step3 - step2: >.2f} seconds')
        return filtered_matches


def token_is_out_of_phrase_range(token: Token, phrase: Phrase, token_searcher: FuzzyTokenSearcher):
    assert token.n in token_searcher.phrase_model.min_token_offset_in_phrase
    assert phrase.phrase_string in token_searcher.phrase_model.min_token_offset_in_phrase[token.n]
    if has_max_start_offset(phrase):
        max_token_offset = token_searcher.phrase_model.max_token_offset_in_phrase[token.n][phrase.phrase_string]
        if token.char_index > phrase.max_start_offset + max_token_offset:
            return True
    if has_max_end_offset(phrase):
        min_token_offset = token_searcher.phrase_model.min_token_offset_in_phrase[token.n][phrase.phrase_string]
        if token.char_index < phrase.max_end_offset + min_token_offset:
            return True
    return False


def get_token_skip_match_type(text_token_string: str, text_token_num_skips: int,
                              skip_matches: SkipMatches, phrase_token_match: str,
                              token_searcher: FuzzyTokenSearcher, debug: int = 0) -> MatchType:
    first = skip_matches.match_skipgrams[phrase_token_match][0]
    last = skip_matches.match_skipgrams[phrase_token_match][-1]
    overlap_start = first.start_offset
    overlap_end = last.start_offset + last.length
    num_skip_matches = len(skip_matches.match_set[phrase_token_match])
    text_token_skip_overlap = num_skip_matches / text_token_num_skips
    phrase_token_skip_overlap = num_skip_matches / token_searcher.token_num_skips[phrase_token_match]
    if text_token_skip_overlap > phrase_token_skip_overlap:
        length_variance = len(text_token_string) - (overlap_end - overlap_start)
    else:
        length_variance = len(phrase_token_match) - (overlap_end - overlap_start)
    if debug > 1:
        print(f"        get_token_skip_match_type - first:", first)
        print(f"        get_token_skip_match_type - last:", last)
        print(f"        get_token_skip_match_type - overlap_start:", overlap_start)
        print(f"        get_token_skip_match_type - overlap_end:", overlap_end)
        print(f"        get_token_skip_match_type - num_skip_matches:", num_skip_matches)
        print(f"        get_token_skip_match_type - text_token_skip_overlap:", text_token_skip_overlap)
        print(f"        get_token_skip_match_type - phrase_token_skip_overlap:", phrase_token_skip_overlap)
        print(f"        get_token_skip_match_type - length_variance:", length_variance)
    if text_token_skip_overlap < token_searcher.config['skipgram_threshold'] and \
            phrase_token_skip_overlap < token_searcher.config['skipgram_threshold']:
        match_type = MatchType.NONE
        if debug > 1:
            print(f"        get_token_skip_match_type - below skipgram thresholds, match_type:", match_type)
    elif length_variance > token_searcher.config['max_token_length_variance']:
        match_type = MatchType.NONE
        if debug > 1:
            print(f"        get_token_skip_match_type - above max length variance, match_type:", match_type)
    elif abs(len(text_token_string) - len(phrase_token_match)) <= token_searcher.config['max_token_length_variance']:
        match_type = MatchType.FULL
        if debug > 1:
            print(f"        get_token_skip_match_type - text and phrase tokens equal length, match_type:", match_type)
    elif len(text_token_string) < len(phrase_token_match):
        match_type = MatchType.PARTIAL_OF_PHRASE_TOKEN
        if debug > 1:
            print(f"        get_token_skip_match_type - phrase token longer than text token, match_type:", match_type)
    else:
        match_type = MatchType.PARTIAL_OF_TEXT_TOKEN
        if debug > 1:
            print(f"        get_token_skip_match_type - text token longer than phrase token, match_type:", match_type)
    return match_type



from unittest import TestCase

from fuzzy_search.search.config import default_config
from fuzzy_search.search.searcher import FuzzySearcher
from fuzzy_search.tokenization.token import Tokenizer


class TestSearcher(TestCase):

    def test_searcher_has_default_config(self):
        searcher = FuzzySearcher()
        for field in default_config:
            with self.subTest(field):
                self.assertEqual(default_config[field], searcher.config[field])

    def test_searcher_has_default_tokenizer(self):
        searcher = FuzzySearcher()
        self.assertEqual(True, isinstance(searcher.tokenizer, Tokenizer))

    def test_searcher_indexes_phrases(self):
        phrase = 'test'
        searcher = FuzzySearcher(phrase_list=[phrase])
        self.assertEqual(1, len(searcher.phrases))

    def test_searcher_indexes_phrase_skipgrams(self):
        phrase = 'test'
        searcher = FuzzySearcher(phrase_list=[phrase])
        self.assertEqual(True, 'te' in searcher.skipgram_index)

    def test_searcher_indexes_phrase_skipgram_phrase(self):
        phrase = 'test'
        searcher = FuzzySearcher(phrase_list=[phrase])
        self.assertEqual(True, phrase in [phrase.phrase_string for phrase in searcher.skipgram_index['te']])

    def test_searcher_passes_tokenizer_to_phrase_model(self):
        phrase = 'test'
        searcher = FuzzySearcher(phrase_list=[phrase])
        self.assertEqual(searcher.tokenizer, searcher.phrase_model.tokenizer)

    def test_searcher_indexes_phrase_tokens(self):
        phrase = 'test'
        searcher = FuzzySearcher(phrase_list=[phrase])
        self.assertEqual(True, phrase in searcher.phrase_model.token_in_phrase)



from typing import List, Union

from .token import Token


class Vocabulary:

    def __init__(self, terms: List[Union[str, Token]] = None, ignorecase: bool = False):
        """A Vocabulary class to map terms to identifiers."""
        self.term_id = {}
        self.id_term = {}
        self.term_freq = {}
        self.ignorecase = ignorecase
        if terms is not None:
            self.add_terms(terms)

    def __repr__(self):
        return f'{self.__class__.__name__}(vocabulary_size="{len(self.term_id)}")'

    def __len__(self):
        return len(self.term_id)

    def reset_index(self):
        self.term_id = {}
        self.id_term = {}
        self.term_freq = {}

    def add_terms(self, terms: List[Union[str, Token]], reset_index: bool = True):
        """Add a list of terms to the vocabulary. Use 'reset_index=True' to reset
        the vocabulary before adding the terms.

        :param terms: a list of terms to add to the vocabulary
        :type terms: List[str]
        :param reset_index: a flag to indicate whether to empty the vocabulary before adding terms
        :type reset_index: bool
        """
        if reset_index is True:
            self.reset_index()
        for term in terms:
            if isinstance(term, Token):
                term = term.n
            term = term.lower() if self.ignorecase else term
            if term in self.term_id:
                continue
            self._index_term(term)

    def has_term(self, term: Union[str, Token], ignorecase: bool = None):
        if ignorecase is None:
            ignorecase = self.ignorecase
        if isinstance(term, Token):
            term = term.n
        term = term.lower() if ignorecase else term
        return True if term in self.term_id else False

    def _index_term(self, term: str):
        term_id = len(self.term_id)
        self.term_id[term] = term_id
        self.id_term[term_id] = term

    def term2id(self, term: str):
        """Return the term ID for a given term."""
        return self.term_id[term] if term in self.term_id else None

    def id2term(self, term_id: int):
        """Return the term for a given term ID."""
        return self.id_term[term_id] if term_id in self.id_term else None



from unittest import TestCase

from fuzzy_search.tokenization.token import Token
from fuzzy_search.tokenization.token import Doc
from fuzzy_search.tokenization.token import Tokenizer
from fuzzy_search.tokenization.token import CustomTokenizer
from fuzzy_search.tokenization.token import tokens2string


class TestToken(TestCase):

    def test_tokenizer_token_has_normalised_string(self):
        token = Token(string='test', index=0, char_index=0)
        self.assertEqual(token.string, token.normalised_string)

    def test_tokenizer_token_can_have_label_string(self):
        token = Token(string='test', index=0, char_index=0, label='test_label')
        self.assertEqual(True, 'test_label' in token.label)

    def test_tokenizer_token_can_have_label_set(self):
        token = Token(string='test', index=0, char_index=0, label={'test', 'label'})
        self.assertEqual(True, 'test' in token.label)

    def test_tokenizer_token_can_check_has_label(self):
        token = Token(string='test', index=0, char_index=0, label={'test', 'label'})
        self.assertEqual(True, token.has_label('label'))

    def test_tokenizer_lower_affects_normalised_string(self):
        token = Token(string='Test', index=0, char_index=0)
        token.lower()
        self.assertEqual('Test', token.string)
        self.assertEqual('test', token.normalised_string)

    def test_tokenizer_update_affects_normalised_string(self):
        token = Token(string='test', index=0, char_index=0)
        updated_token = token.update('text')
        self.assertEqual('text', updated_token.normalised_string)
        self.assertEqual('test', updated_token.string)

    def test_tokenizer_update_returns_new_token(self):
        token = Token(string='test', index=0, char_index=0)
        updated_token = token.update('text')
        self.assertEqual('test', token.normalised_string)
        self.assertEqual('text', updated_token.normalised_string)


class TestTokenizer(TestToken):

    def setUp(self) -> None:
        self.text = 'This is an example sentence.'

    def test_tokenize_returns_a_list_of_tokens(self):
        tokenizer = Tokenizer(include_boundary_tokens=False)
        tokens = tokenizer.tokenize(self.text)
        for ti, token in enumerate(tokens):
            with self.subTest(ti):
                self.assertEqual(True, isinstance(token, Token))

    def test_tokenize_doc_returns_a_document_object(self):
        tokenizer = Tokenizer(include_boundary_tokens=False)
        doc = tokenizer.tokenize_doc(self.text)
        self.assertEqual(True, isinstance(doc, Doc))

    def test_tokenizer_tracks_char_index(self):
        tokenizer = Tokenizer(include_boundary_tokens=False)
        tokens = tokenizer.tokenize(self.text)
        token = tokens[3]
        self.assertEqual(self.text.index(token.t), token.char_index)

    def test_tokenizer_defaults_to_no_boundary_tokens(self):
        tokenizer = Tokenizer(include_boundary_tokens=False)
        tokens = tokenizer.tokenize(self.text)
        self.assertEqual(False, tokens[0].string == '<START>')

    def test_tokenizer_can_add_boundaries(self):
        tokenizer = Tokenizer(include_boundary_tokens=True)
        tokens = tokenizer.tokenize(self.text)
        self.assertEqual(True, tokens[0].string == '<START>')

    def test_tokenizer_ignorecase_keeps_original_case_in_string(self):
        tokenizer = Tokenizer(ignorecase=True)
        tokens = tokenizer.tokenize(self.text)
        self.assertEqual(True, tokens[0].string == 'This')

    def test_tokenizer_ignorecase_lowercases_normalised_string(self):
        tokenizer = Tokenizer(ignorecase=True)
        tokens = tokenizer.tokenize(self.text)
        self.assertEqual(True, tokens[0].normalised_string == 'this')


class TestCustomTokenizer(TestCase):

    def setUp(self) -> None:
        self.text = 'This is an example sentence.'

    def test_any_string_tokenizing_func_returns_doc(self):
        def tokenize_func(text):
            return text.split(' ')
        tokenizer = CustomTokenizer(tokenizer_func=tokenize_func)
        tokens = tokenize_func(self.text)
        custom_tokens = tokenizer.tokenize(self.text)
        for ti, token in enumerate(tokens):
            with self.subTest(ti):
                self.assertEqual(token, custom_tokens[ti].t)


class TestToken2String(TestCase):

    def setUp(self) -> None:
        self.string = "Yeah, well, you know, that's just like, uh, your opinion, man."
        self.tokenizer = Tokenizer()
        self.tokens = self.tokenizer.tokenize(self.string)

    def test_text2string_can_reconstruct_original_string(self):
        new_string = tokens2string(self.tokens)
        self.assertEqual(new_string, self.string)




from unittest import TestCase

from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.phrase.phrase_model import PhraseModel
from fuzzy_search.search.phrase_searcher import FuzzyPhraseSearcher
from fuzzy_search.match.skip_match import filter_skipgram_threshold
from fuzzy_search.match.skip_match import get_skipmatch_candidates
from fuzzy_search.tokenization.token import Tokenizer


class TestFuzzyPhraseSearcher(TestCase):

    def setUp(self) -> None:
        self.searcher = FuzzyPhraseSearcher()

    def test_can_make_default_phrase_searcher(self):
        self.assertNotEqual(self.searcher, None)

    def test_can_add_phrases_as_strings(self):
        phrase = "test"
        self.searcher.index_phrases(phrases=[phrase])
        phrase_object = self.searcher.phrases.pop()
        self.assertEqual(phrase_object.phrase_string, phrase)

    def test_can_add_phrases_as_phrase_objects(self):
        phrase = Phrase("test")
        self.searcher.index_phrases(phrases=[phrase])
        self.assertTrue(phrase in self.searcher.phrases)

    def test_can_generate_skip_matches(self):
        phrase = "test"
        self.searcher.index_phrases(phrases=[phrase])
        text = "this is a test"
        skip_matches = self.searcher.find_skipgram_matches({"text": text})
        phrase_object = self.searcher.phrases.pop()
        self.assertTrue(phrase_object in skip_matches.match_set)

    def test_can_filter_skipgram_threshold(self):
        phrase = Phrase("test")
        self.searcher.index_phrases(phrases=[phrase])
        text = "this is a test"
        skip_matches = self.searcher.find_skipgram_matches({"text": text})
        phrases = filter_skipgram_threshold(skip_matches, 0.5)
        self.assertEqual(len(phrases), 1)

    def test_can_get_candidates(self):
        phrase_model = PhraseModel(phrases=["test"])
        self.searcher.index_phrase_model(phrase_model=phrase_model)
        text = {"text": "this is a test"}
        skip_matches = self.searcher.find_skipgram_matches(text)
        phrases = get_skipmatch_candidates(text, skip_matches, 0.5, phrase_model=phrase_model)
        self.assertEqual(len(phrases), 1)

    def test_finds_multiple_candidates(self):
        phrase_model = PhraseModel(phrases=["test"])
        self.searcher.index_phrase_model(phrase_model=phrase_model)
        text = {"text": "a test is a test is a test"}
        skip_matches = self.searcher.find_skipgram_matches(text)
        phrases = get_skipmatch_candidates(text, skip_matches, 0.5, phrase_model=phrase_model)
        self.assertEqual(len(phrases), 3)

    def test_searcher_finds_near_match(self):
        phrase = "contains"
        self.searcher.index_phrases(phrases=[phrase])
        text = "This text consaint some typos."
        matches = self.searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_is_case_sensitive(self):
        phrase = "contains"
        self.searcher.index_phrases(phrases=[phrase])
        text = "This text CONSAINT some typos."
        matches = self.searcher.find_matches(text)
        self.assertEqual(len(matches), 0)

    def test_searcher_handles_ignorecase(self):
        searcher = FuzzyPhraseSearcher(config={"ignorecase": True})
        phrase = "contains"
        searcher.index_phrases(phrases=[phrase])
        text = "This text CONSAINT some typos."
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_uses_word_boundaries(self):
        phrase = "contains"
        self.searcher.index_phrases(phrases=[phrase])
        text = "This text containsi some typos."
        matches = self.searcher.find_matches(text)
        self.assertEqual(isinstance(matches, list), True)
        self.assertEqual("containsi", matches[0].string)

    def test_searcher_finds_repeat_phrases_as_multiple_matches(self):
        phrase = "contains"
        self.searcher.index_phrases(phrases=[phrase])
        text = "This text contains contains some repetition."
        matches = self.searcher.find_matches(text)
        self.assertEqual(len(matches), 2)
        self.assertEqual("contains", matches[0].string)
        self.assertEqual("contains", matches[1].string)

    def test_searcher_finds_correct_start(self):
        phrase = "contains"
        self.searcher.index_phrases(phrases=[phrase])
        text = "This text con contains some weirdness."
        matches = self.searcher.find_matches(text)
        self.assertEqual("contains", matches[0].string)

    def test_searcher_allows_length_variance(self):
        phrase = "coffee"
        self.searcher.index_phrases(phrases=[phrase])
        text = "For sale two units of coffy."
        matches = self.searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_allows_length_variance_2(self):
        phrase = "Makelaars"
        self.searcher.index_phrases(phrases=[phrase])
        text = 'door de Alakei&ers by na gecompletecrt'
        matches = self.searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_accepts_phrases_on_init(self):
        phrase = "Makelaars"
        searcher = FuzzyPhraseSearcher(phrase_list=[phrase])
        self.assertEqual(1, len(searcher.phrases))

    def test_searcher_accepts_tokenized_document(self):
        phrase = "Makelaars"
        self.searcher.index_phrases(phrases=[phrase])
        text = 'door de Alakei&ers by na gecompletecrt'
        tokenizer = Tokenizer()
        doc = tokenizer.tokenize_doc(text)
        error = None
        try:
            matches = self.searcher.find_matches(doc)
        except TypeError as err:
            error = err
        self.assertEqual(None, error)
        self.assertEqual(1, len(matches))


class TestFuzzySearchVariants(TestCase):

    def setUp(self) -> None:
        self.searcher = FuzzyPhraseSearcher(config={"include_variants": True})
        self.phrase = {"phrase": "okay", "variants": ["OK"]}

    def test_searcher_can_toggle_variants(self):
        self.assertEqual(self.searcher.include_variants, True)

    def test_searcher_can_register_variants(self):
        self.searcher.index_phrase_model(phrase_model=PhraseModel([self.phrase]))
        self.assertEqual(len(self.searcher.variants), 1)
        variant = self.searcher.variants.pop()
        self.assertEqual(variant.phrase_string, "OK")

    def test_searcher_can_match_variants(self):
        self.searcher.index_phrase_model(phrase_model=PhraseModel([self.phrase]))
        text = "This text is okay and this test is OK."
        matches = self.searcher.find_matches(text, include_variants=True)
        self.assertEqual(matches[1].phrase.phrase_string, self.phrase["phrase"])
        self.assertEqual(matches[1].variant.phrase_string, self.phrase["variants"][0])


class TestFuzzySearchDistractors(TestCase):

    def setUp(self) -> None:
        self.searcher = FuzzyPhraseSearcher(config={"filter_distractors": True})
        self.phrase = {"phrase": "okay", "distractors": ["OK"]}

    def test_searcher_can_toggle_distractors(self):
        self.assertEqual(self.searcher.filter_distractors, True)

    def test_searcher_can_register_distractors(self):
        self.searcher.index_phrase_model(phrase_model=PhraseModel([self.phrase]))
        self.assertEqual(len(self.searcher.distractors), 1)
        distractor = self.searcher.distractors.pop()
        self.assertEqual(distractor.phrase_string, "OK")

    def test_searcher_can_match_distractors(self):
        phrase = {"phrase": "baking", "distractors": ["braking"]}
        self.searcher.index_phrase_model(phrase_model=PhraseModel([phrase]))
        text = "This text is about baking and not about braking."
        matches = self.searcher.find_matches(text, filter_distractors=True)
        self.assertEqual(len(matches), 1)


class TestFuzzySearchExactMatch(TestCase):

    def setUp(self) -> None:
        self.searcher = FuzzyPhraseSearcher()
        self.phrase = {"phrase": "baking", "distractors": ["braking"]}
        self.searcher.index_phrase_model(phrase_model=PhraseModel([self.phrase]))

    def test_fuzzy_search_can_search_exact_match(self):
        text = "This text is about baking and not about braking."
        matches = self.searcher.find_exact_matches(text, debug=True)
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].string, "baking")

    def test_fuzzy_search_can_search_exact_match_with_word_boundaries(self):
        text = "This text is about baking and not about rebaking."
        matches = self.searcher.find_exact_matches(text, use_word_boundaries=True)
        self.assertEqual(len(matches), 1)

    def test_fuzzy_search_can_search_exact_match_without_word_boundaries(self):
        text = "This text is about baking and not about rebaking."
        matches = self.searcher.find_exact_matches(text, use_word_boundaries=False)
        self.assertEqual(len(matches), 2)
        self.assertEqual(matches[1].string, "baking")

    def test_fuzzy_search_can_search_exact_match_with_special_characters(self):
        searcher = FuzzyPhraseSearcher()
        phrase = {"phrase": "[baking]", "distractors": ["braking"]}
        searcher.index_phrase_model(phrase_model=PhraseModel([phrase]))
        text = "This text is about [baking] and not about braking."
        matches = searcher.find_exact_matches(text, use_word_boundaries=False)
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0].string, "[baking]")

    def test_text_split(self):
        text = {
            "text": "Ntfangen een Missive van den Gouverneur Generaal van het eiland Amoras, "
                    + "verfoekende, dat dit beter getest moet worden.",
            "id": "urn:republic/inv=3825/meeting=1775-01-28/para=2"
        }
        phrases = [
            {"phrase": "ONtfangen een Missive van"},
            {"phrase": "Missive"},
            {"phrase": "Gouverneur Generaal"},
            {"phrase": "Gouverneur"},
            {"phrase": "Generaal van de"},
            {"phrase": "versoekende"}
        ]
        phrase_model = PhraseModel(model=phrases)
        searcher = FuzzyPhraseSearcher()
        searcher.index_phrase_model(phrase_model)
        exact_matches = searcher.find_exact_matches(text)
        self.assertEqual(len(exact_matches), 3)


class TestSearcherRealData(TestCase):

    def setUp(self) -> None:
        self.text1 = "ie Veucris den 5. Januaris 1725. PR&ASIDE, Den Heere Bentinck. PRASENTIEBUS, " + \
                     "De Heeren Jan Welderen , van Dam, Torck , met een extraordinaris Gedeputeerde" + \
                     " uyt de Provincie van Gelderlandt. Van Maasdam , vanden Boeizelaar , Raadtpen" + \
                     "fionaris van Hoornbeeck , met een extraordinaris Gedeputeerde uyt de Provinci" + \
                     "e van Hollandt ende Welt-Vrieslandt. Velters, Ockere , Noey; van Hoorn , met " + \
                     "een extraordinaris Gedeputeerde uyt de Provincie van Zeelandt. Van Renswoude " + \
                     ", van Voor{t. Van Schwartzenbergh, vander Waayen, Vegilin Van I{elmuden. Van " + \
                     "Iddekinge ‚ van Tamminga."
        self.text2 = "Mercuri: den 10. Jangarii, 1725. ia PRESIDE, Den Heere an Iddekinge. PRA&SENT" + \
                     "IBUS, De Heeren /an Welderen , van Dam, van Wynbergen, Torck, met een extraor" + \
                     "dinaris Gedeputeerde uyt de Provincie van Gelderland. Van Maasdam , Raadtpenf" + \
                     "ionaris van Hoorn=beeck. Velters, Ockerfe, Noey. Taats van Amerongen, van Ren" + \
                     "swoude. Vander Waasen , Vegilin, ’ Bentinck, van I(elmaden. Van Tamminga."
        self.config = {
            "char_match_threshold": 0.6,
            "ngram_threshold": 0.5,
            "levenshtein_threshold": 0.6,
            "ignorecase": False,
            "max_length_variance": 3,
            "ngram_size": 2,
            "skip_size": 2,
        }
        self.searcher = FuzzyPhraseSearcher(config=self.config)
        # create a list of domain phrases
        self.domain_phrases = [
            # terms for the chair and attendants of a meeting
            "PRAESIDE",
            "PRAESENTIBUS",
            # some weekdays in Latin
            "Veneris",
            "Mercurii",
            # some date phrase where any date in January 1725 should match
            "den .. Januarii 1725"
        ]
        self.phrase_model = PhraseModel(phrases=self.domain_phrases)
        # register the keywords with the searcher
        self.searcher.index_phrase_model(self.phrase_model)

    def test_fuzzy_search_text1_finds_four_matches(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(len(matches), 4)

    def test_fuzzy_search_text1_finds_friday(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[0].string, "Veucris")

    def test_fuzzy_search_text1_finds_date(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[1].string, "den 5. Januaris 1725")

    def test_fuzzy_search_text1_finds_president(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[2].string, "PR&ASIDE")

    def test_fuzzy_search_text1_finds_attendants(self):
        matches = self.searcher.find_matches(self.text1)
        self.assertEqual(matches[3].string, "PRASENTIEBUS")

    def test_fuzzy_search_text2_finds_four_matches(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(len(matches), 4)

    def test_fuzzy_search_text2_finds_friday(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[0].string, "Mercuri")

    def test_fuzzy_search_text2_finds_date(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[1].string, "den 10. Jangarii, 1725")

    def test_fuzzy_search_text2_finds_president(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[2].string, "PRESIDE")

    def test_fuzzy_search_text2_finds_attendants(self):
        matches = self.searcher.find_matches(self.text2)
        self.assertEqual(matches[3].string, "PRA&SENTIBUS")


class TestSearcherRealData2(TestCase):

    def setUp(self) -> None:
        self.text1 = 'TS gehoort het rapport van de Heeren I van Lynden'
        self.config = {
            "char_match_threshold": 0.6,
            "ngram_threshold": 0.5,
            "levenshtein_threshold": 0.6,
            "ignorecase": False,
            "include_variants": True,
            "max_length_variance": 3,
            "ngram_size": 2,
            "skip_size": 2,
        }
        self.searcher = FuzzyPhraseSearcher(config=self.config)
        # create a list of domain phrases
        self.domain_phrases = [
            {
                'phrase': 'den Heere',
                'variants': [
                    'de Heer',
                    'de Heeren',
                ]
            }
        ]
        self.phrase_model = PhraseModel(phrases=self.domain_phrases)
        # register the keywords with the searcher
        self.searcher.index_phrase_model(self.phrase_model)

    def test_searcher_find_no_overlapping_variants(self):
        phrase_matches = self.searcher.find_matches(self.text1)
        self.assertEqual(len(phrase_matches), 1)

    def test_searcher_allows_length_variance_2(self):
        searcher = FuzzyPhraseSearcher(config=self.config)
        searcher.ignorecase = True
        phrase = "Admiraliteiten in t gemeen"
        searcher.index_phrases(phrases=[phrase])
        text = 'aaniraliteyten in het gemeen'
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_allows_length_variance_3(self):
        searcher = FuzzyPhraseSearcher(config=self.config)
        searcher.ignorecase = True
        phrase = 'Admiraliteit in Vriesland'
        searcher.index_phrases(phrases=[phrase])
        text = 'AduiraliteytVrieslaidt'
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_finds_DONtfangen(self):
        searcher = FuzzyPhraseSearcher(config=self.config)
        searcher.ignorecase = True
        phrase = "ONtfangen een Missive van"
        searcher.index_phrases(phrases=[phrase])
        text = 'DONtfangen een Missive van den Heere vander Goes'
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 1)

    def test_searcher_finds_long_opening(self):
        searcher = FuzzyPhraseSearcher(config=self.config)
        searcher.ignorecase = True
        phrases = [
            "hebben ter Vergaderinge ingebraght",
            'hebben ter Vergaderinge ingebragt en laaten leezen de Resolutie'
        ]
        searcher.index_phrases(phrases=phrases)
        text = "De Heeren Gedeputeerden van de Provincie van Zeelandt, hebben ter Vergaderinge ingebraght en laten lesen de Resolutie van de Heeren Staten van de hoogh-gemelde Provincie hare Principalen, raeckende het negotieren van hare quote voor een derde part in de Petitie tot de extraordinaris Equipage voor het loopende jaer, volgende de voorschreve Resolutie hier na geinsereert."
        matches = searcher.find_matches(text)
        self.assertEqual(len(matches), 2)




from unittest import TestCase

from fuzzy_search.phrase.phrase import Phrase
from fuzzy_search.tokenization.string import SkipGram
from fuzzy_search.search.phrase_searcher import Candidate


class TestCandidate(TestCase):

    def test_candidate_detects_no_match_with_no_skip_match(self):
        phrase = Phrase('test')
        candidate = Candidate(phrase)
        self.assertEqual(candidate.is_match(0.5), False)

    def test_candidate_detects_no_match(self):
        phrase = Phrase('test')
        candidate = Candidate(phrase)
        skipgram = SkipGram('ts', 0, len(phrase.phrase_string), 3)
        candidate.add_skip_match(skipgram)
        self.assertEqual(candidate.is_match(0.5), False)

    def test_candidate_has_skipgram_overlap(self):
        phrase = Phrase('test')
        candidate = Candidate(phrase)
        skipgram = SkipGram('ts', 0, len(phrase.phrase_string), 3)
        candidate.add_skip_match(skipgram)
        self.assertTrue(candidate.get_skip_set_overlap() > 0.0)


